;------------------------------------------------------------------------
; The BIOS Move Block function requires that the caller supply a 48-byte
; Global Descriptor table, consisting of 6 segment descriptors.
;------------------------------------------------------------------------

SEGMENT_DESCRIPTOR STRUCT
  seg_limit     WORD  0
  base_low      WORD  0
  base_mid      BYTE  0
  access_rights BYTE  0
  extra_byte    BYTE  0
  base_high     BYTE  0
SEGMENT_DESCRIPTOR ENDS

;----------------------------------------------------------------------
; The BIOS will take care of initializing the dummy, gdtptr, code, and
; stack descriptors. Here we specify initializers for the source and
; dest descriptors with a maximum segment limit and page granularity
; (so the segments can potentially cover the entire 32-bit address
; space), and appropriate access rights. The segment base will need to
; be initialized at run time.
;----------------------------------------------------------------------

MOVE_BLOCK_GDT STRUCT
  dummy       SEGMENT_DESCRIPTOR <>
  gdtptr      SEGMENT_DESCRIPTOR <>
  source      SEGMENT_DESCRIPTOR <0ffffh, , , 93h, 8fh>
  dest        SEGMENT_DESCRIPTOR <0ffffh, , , 93h, 8fh>
  code        SEGMENT_DESCRIPTOR <>
  stack       SEGMENT_DESCRIPTOR <>
MOVE_BLOCK_GDT ENDS

;=========================================================================
.model small, c
.386
include support.asm
.stack
.data
    gdt   MOVE_BLOCK_GDT <>
    buff  db 16 dup(0)
.code
;=========================================================================

;-------------------------------------------------------------------------
; This procedure copies a block from extended memory to real-mode memory.
; Returns zero on success, or a non-zero BIOS error code on failure.
;-------------------------------------------------------------------------

GetXmBlock proc uses ecx edx si es destSeg:WORD,
                                   destOffset:WORD,
                                   source:DWORD,
                                   cb:WORD

    mov edx, source
    mov gdt.source.base_low, dx
    shr edx, 16
    mov gdt.source.base_mid, dl
    shr edx, 8
    mov gdt.source.base_high, dl

    ;print hexbyte$(gdt.source.base_high)
    ;print hexbyte$(gdt.source.base_mid)
    ;print hexword$(gdt.source.base_low),NL

    ;print NL,hexword$(destSeg),NL
    ;print hexword$(destOffset),NL

    movzx edx, destSeg
    shl edx, 4
    movzx ecx, destOffset
    add edx, ecx
    mov gdt.dest.base_low, dx
    shr edx, 16
    mov gdt.dest.base_mid, dl
    shr edx, 8
    mov gdt.dest.base_high, dl

    ;print hexbyte$(gdt.dest.base_high)
    ;print hexbyte$(gdt.dest.base_mid)
    ;print hexword$(gdt.dest.base_low),NL

    ;----------------------------------------
    ; The function expects CX to contain the
    ; number of 16-bit words to move.
    ;----------------------------------------

    mov cx, cb
    shr cx, 1

    push seg gdt
    pop es
    mov si, OFFSET gdt

    mov ah, 87h
    int 15h
    jnc @F
    movzx ax, ah
    ret

  @@:
    xor ax, ax
    ret

GetXmBlock endp

;=========================================================================

;-------------------------------------------------------------------------
; This procedure copies a block from real-mode memory to extended memory.
; Returns zero on success, or a non-zero BIOS error code on failure.
;-------------------------------------------------------------------------

PutXmBlock proc uses ecx edx si es dest:DWORD,
                                   sourceSeg:WORD,
                                   sourceOffset:WORD,
                                   cb:WORD

    movzx edx, sourceSeg
    shl edx, 4
    movzx ecx, sourceOffset
    add edx, ecx
    mov gdt.source.base_low, dx
    shr edx, 16
    mov gdt.source.base_mid, dl
    shr edx, 8
    mov gdt.source.base_high, dl

    ;print NL,hexword$(sourceSeg),NL
    ;print hexword$(sourceOffset),NL

    ;print hexbyte$(gdt.source.base_high)
    ;print hexbyte$(gdt.source.base_mid)
    ;print hexword$(gdt.source.base_low),NL

    mov edx, dest
    mov gdt.dest.base_low, dx
    shr edx, 16
    mov gdt.dest.base_mid, dl
    shr edx, 8
    mov gdt.dest.base_high, dl

    ;print hexbyte$(gdt.dest.base_high)
    ;print hexbyte$(gdt.dest.base_mid)
    ;print hexword$(gdt.dest.base_low),NL

    ;----------------------------------------
    ; The function expects CX to contain the
    ; number of 16-bit words to move.
    ;----------------------------------------

    mov cx, cb
    shr cx, 1

    push seg gdt
    pop es
    mov si, OFFSET gdt

    mov ah, 87h
    int 15h
    jnc @F
    movzx ax, ah
    ret

  @@:
    xor ax, ax
    ret

PutXmBlock endp
;=========================================================================
.startup
;=========================================================================

    invoke GetXmBlock, SEG buff, OFFSET buff, 0fffffff0h, 16

    print word$(ax),NL

    xor ecx, ecx
    mov ebx, OFFSET buff
    .WHILE ecx < 16
      mov al, [ebx+ecx]
      print hexbyte$(al)
      inc ecx
    .ENDW

    print NL,NL

    invoke PutXmBlock, 0fffffff0h, SEG buff, OFFSET buff, 16

    print word$(ax),NL,NL

    xor ah, ah
    int 16h

;=========================================================================
.exit
end

